/*=========================================================================

  Program:   Visualization Toolkit
  Module:    $RCSfile$

  Copyright (c) Ken Martin, Will Schroeder, Bill Lorensen
  All rights reserved.
  See Copyright.txt or http://www.kitware.com/Copyright.htm for details.

     This software is distributed WITHOUT ANY WARRANTY; without even
     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
     PURPOSE.  See the above copyright notice for more information.

=========================================================================*/
// This plugin has been developed and contributed by Sven Buijssen, TU
// Dortmund, Germany.
// Thanks to Bryn Lloyd (blloyd@vision.ee.ethz.ch) at ETH Zuerich for
// developing and sharing vtkTensorGlyphFilter, the ancestor of this
// filter. That filter's output (i.e. spheres) can be mimicked by setting both
// ThetaRoundness and PhiRoundness to 1.0.
// Thanks to Gordon Kindlmann for pointing out at VisSym04 that superquadrics
// can be put to a good use to visualize tensors and for pointing out to me
// an insufficient initial implementation of this filter.

#include "vtkSuperquadricTensorGlyphFilter.h"

#include "vtkInformation.h"
#include "vtkInformationVector.h"
#include "vtkObjectFactory.h"
#include "vtkPointData.h"
#include "vtkPolyData.h"
#include "vtkTensorGlyphSameEigensystem.h"
#include "vtkSuperquadricSource2.h"
#include "vtkCellArray.h"
#include "vtkAppendPolyData.h"
#include "vtkMath.h"
#include "vtkDoubleArray.h"
#include "vtkIdTypeArray.h"


vtkStandardNewMacro(vtkSuperquadricTensorGlyphFilter);

vtkSuperquadricTensorGlyphFilter::vtkSuperquadricTensorGlyphFilter()
{
  this->SetNumberOfInputPorts(1);
  this->ThetaResolution = 16;
  this->PhiResolution = 16;
  this->ThetaRoundness = 0.3;
  this->PhiRoundness = 0.3;
  this->Gamma = 1.5;
  this->ScaleFactor = 1.0;
  this->ExtractEigenvalues = 0;

  // by default, process active point tensors
  this->SetInputArrayToProcess(0, 0, 0, vtkDataObject::FIELD_ASSOCIATION_POINTS,
                               vtkDataSetAttributes::TENSORS);

  // by default, process active point scalars
  this->SetInputArrayToProcess(1, 0, 0, vtkDataObject::FIELD_ASSOCIATION_POINTS,
                               vtkDataSetAttributes::SCALARS);
}

vtkSuperquadricTensorGlyphFilter::~vtkSuperquadricTensorGlyphFilter()
{
}

int vtkSuperquadricTensorGlyphFilter::FillInputPortInformation(
  int vtkNotUsed(port), vtkInformation* info)
{
  info->Set(vtkAlgorithm::INPUT_REQUIRED_DATA_TYPE(), "vtkDataSet");
  return 1;
}


int vtkSuperquadricTensorGlyphFilter::RequestData(
  vtkInformation *vtkNotUsed(request),
  vtkInformationVector **inputVector,
  vtkInformationVector *outputVector)
{
  // get the info objects
  vtkInformation *inInfo = inputVector[0]->GetInformationObject(0);
  vtkInformation *outInfo = outputVector->GetInformationObject(0);

  // get the input and output
  vtkDataSet *input = vtkDataSet::SafeDownCast(inInfo->Get(vtkDataObject::DATA_OBJECT()));
  vtkPolyData *output = vtkPolyData::SafeDownCast(outInfo->Get(vtkDataObject::DATA_OBJECT()));

  vtkPointData *inputPD;
  vtkPointData *outputPD;
  vtkSuperquadricSource2 *superquadric;
  vtkAppendPolyData* append;
  vtkIdType numPtsIn, numGlyphPtsTotal, ptIncr, inPtId, i;
  int j, axis, abort = 0;
  vtkTensorGlyphSameEigensystem* tensors = NULL;
  vtkDataArray *inTensors = NULL;
  vtkDataArray *inScalars = NULL;
  double tensor[9];
  double *m[3], w[3], *v[3];
  double m0[3], m1[3], m2[3];
  double v0[3], v1[3], v2[3];
  vtkDoubleArray *lambda;
  vtkIdTypeArray *numGlyphPts;
  double sum, cl, cp, maxScale, theta, phi; // cs
  vtkIdType id[1];
  int showWarning = 0;

  // set up working matrices
  m[0] = m0; m[1] = m1; m[2] = m2;
  v[0] = v0; v[1] = v1; v[2] = v2;

  inputPD = input->GetPointData();
  if (!inputPD)
    {
    vtkErrorMacro(<<"No data to glyph!");
    return 1;
    }

  numPtsIn = input->GetNumberOfPoints();
  if (numPtsIn < 1)
    {
    vtkErrorMacro(<<"No points to glyph!");
    return 1;
    }

  // Store eigenvalues for later useas data array
  if ( this->ColorGlyphs &&
       this->ColorMode == COLOR_BY_EIGENVALUES)
    {
    lambda = vtkDoubleArray::New();
    lambda->SetNumberOfComponents(3);
    lambda->SetNumberOfTuples(numPtsIn);
    lambda->SetName("Eigenvalues");
    }

  inScalars = this->GetInputArrayToProcess(1, inputVector);
  inTensors = this->GetInputArrayToProcess(0, inputVector);
  if (!inTensors)
    {
    vtkErrorMacro(<<"No data to glyph!");
    return 1;
    }

  append = vtkAppendPolyData::New();

  // Not every input point necessarily has valid tensor data and is
  // hence glyphed; additionally, not every glyph necessarily has the
  // same number of points.
  numGlyphPts = vtkIdTypeArray::New();
  numGlyphPts->SetNumberOfValues(numPtsIn);
  for (inPtId = 0; inPtId < numPtsIn && !abort; inPtId++)
    {
    if (! (inPtId % 500))
      {
      // this->UpdateProgress(static_cast<double>(inPtId)/numPtsIn);
      // abort = this->GetAbortExecute();
      }

    if (!abort)
      {
      inTensors->GetTuple(inPtId, tensor);

      // Init eigenvalues
      w[0] = w[1] = w[2] = 0.0;

      // If trace is 0, there is no need to continue as no glyph would
      // be created anyway.
      sum = tensor[0] + tensor[4] + tensor[8];
      if (fabs(sum) <= VTK_DBL_EPSILON)
        {
        numGlyphPts->SetValue(inPtId,0);
        }
      else if (sum < 0)
        {
        numGlyphPts->SetValue(inPtId,0);
        showWarning = 1;
        }
      else
        {
        // Copy point inPtId and its point data to a single point input object
        vtkPolyData* singlePoint = vtkPolyData::New();
        vtkPoints* newPt = vtkPoints::New();
        vtkCellArray* newVertex = vtkCellArray::New();
        newPt->SetNumberOfPoints(1);
        newPt->SetPoint(0,input->GetPoint(inPtId));
        singlePoint->SetPoints(newPt);
        newPt->Delete();
        singlePoint->GetPointData()->CopyAllOn();
        singlePoint->GetPointData()->CopyAllocate(inputPD, 1);
        singlePoint->GetPointData()->CopyData(inputPD, inPtId, 0);

        // Create vertex for this point
        id[0] = inPtId;
        newVertex->InsertNextCell(1,id);
        singlePoint->SetVerts(newVertex);

        // set the active tensor for next operation
        singlePoint->GetPointData()->SetActiveTensors( inTensors->GetName() );
        if (inScalars)
          {
          singlePoint->GetPointData()->SetActiveScalars( inScalars->GetName() );
          }

        tensors = vtkTensorGlyphSameEigensystem::New();
        tensors->SetInput(singlePoint);
        tensors->SetExtractEigenvalues(ExtractEigenvalues);
        tensors->SetColorGlyphs(this->GetColorGlyphs());
        tensors->SetColorMode(this->GetColorMode());
        tensors->SetScaleFactor(ScaleFactor);
        tensors->SetClampScaling(ClampScaling);
        tensors->SetMaxScaleFactor(MaxScaleFactor);
        tensors->SetThreeGlyphs(0);
        singlePoint->Delete();

        superquadric = vtkSuperquadricSource2::New();
        superquadric->SetThetaResolution(this->ThetaResolution);
        superquadric->SetPhiResolution(this->PhiResolution);
        superquadric->ToroidalOff();
        axis = 2;  // superquadrics with rotational symmetriy around the z axis

        // Set roundness per tensor, if requested
        theta = this->ThetaRoundness;
        phi = this->PhiRoundness;

        if (!this->FixedThetaPhiRoundness)
          {
          theta = 0.0;
          phi = 0.0;

          for (j=0; j<3; j++)
            {
            for (i=0; i<3; i++)
              {
              m[i][j] = tensor[i+3*j];
              }
            }
          // This is a costly operation, one that is repeated again in
          // vtkTensorGlyph if ExtractEigenvalues is set. To optimize speed this
          // information is passed to a custom subclass of vtkTensorGlyph.
          vtkMath::Jacobi(m, w, v);
          // eigenvalues are stored in w, sorted largest to smallest

          // avoid calculating them again in vtkTensorGlyph:
          tensors->SetEigenvectorsEigenValues(v, w);

          // Calculation according to equation (7) from
          // Gordon Kindlmann, Superquadric Tensor Glyphs, Joint
          // EUROGRAPHICS - IEEE TCVG Symposium on Visualization (2004),
          // http://www.cs.utah.edu/~gk/papers/vissym04/vissym04kindlmann.pdf
          // The algorithm assumes three non-negative eigenvalues, though!
          //
          // Given that they are sorted largest to smallest, it suffices to check
          // whether the smallest is still non-negative.
          if (w[2] > 0.)
            {
            sum = w[0] + w[1] + w[2];
            cl = (w[0] - w[1]) / sum;
            cp = 2*(w[1] - w[2]) / sum;
            //cs = 3*w[2] / sum;

            if (cl >= cp)
              {
              axis  = 0;  // superquadric with rotational symmetriy around the x axis
              theta = pow(1 - cp, this->Gamma);
              phi   = pow(1 - cl, this->Gamma);
              }
            else
              {
              axis  = 2;  // superquadric with rotational symmetriy around the z axis
              theta = pow(1 - cl, this->Gamma);
              phi   = pow(1 - cp, this->Gamma);
              }
            }
          else
            {
            vtkErrorMacro(<<"Point " << inPtId << " has at least one negative tensor eigenvalue. Superquadric tensor glyphing not feasible.");
            superquadric->Delete();
            tensors->Delete();
            continue;
            }
          }
        superquadric->SetAxisOfSymmetry(axis);
        superquadric->SetThetaRoundness(theta);
        superquadric->SetPhiRoundness(phi);
        tensors->SetSource(superquadric->GetOutput());
        tensors->SetScaleFactor(this->ScaleFactor);
        tensors->Update();
        numGlyphPts->SetValue(inPtId, tensors->GetOutput()->GetNumberOfPoints());
        superquadric->Delete();

        // Add tensor glyph filter's output to output
        append->AddInput( vtkPolyData::SafeDownCast( tensors->GetOutput()) );
        append->Update();
        tensors->Delete();
        }
      }

    if ( this->ColorGlyphs &&
         this->ColorMode == COLOR_BY_EIGENVALUES)
      {
      maxScale = 1.0;
      if ( this->ClampScaling )
        {
        for (maxScale=0.0, i=0; i<3; i++)
          {
          if ( maxScale < fabs(w[i]) )
            {
            maxScale = fabs(w[i]);
            }
          }
        if ( maxScale > this->MaxScaleFactor )
          {
          maxScale = this->MaxScaleFactor / maxScale;
          }
        }
      for (i=0; i<3; i++)
        {
        lambda->SetComponent(inPtId, i, w[i] * maxScale);
        }
      }
    }

  if (append)
    output->ShallowCopy(append->GetOutput());

  if (showWarning)
    {
    vtkWarningMacro(<< "Tensors with non-positive eigenvalues have been omitted. This "
                    << "filter assumes three real, positive eigenvalues.");
    }

  outputPD = output->GetPointData();

  // In constrast to vtkTensorGlyph vtkTensorGlyphSameEigensystem names the
  // data fields it creates such that ParaView will color them. No need
  // to compensate here.

  numGlyphPtsTotal = 0;
  for (inPtId = 0; inPtId < numPtsIn; inPtId++)
    {
      numGlyphPtsTotal += numGlyphPts->GetValue(inPtId);
    }

  // Add eigenvalues as data array
  if ( this->ColorGlyphs &&
       this->ColorMode == COLOR_BY_EIGENVALUES)
    {
    vtkDoubleArray *eigenvalues = vtkDoubleArray::New();
    eigenvalues->SetNumberOfComponents(3);
    eigenvalues->SetNumberOfTuples(numGlyphPtsTotal);
    eigenvalues->SetName(lambda->GetName());
    ptIncr = 0;
    for (inPtId=0; inPtId < numPtsIn; inPtId++)
      {
      for (i=0; i < numGlyphPts->GetValue(inPtId); i++)
        {
        eigenvalues->SetTuple(ptIncr+i, inPtId, lambda);
        }
      ptIncr += numGlyphPts->GetValue(inPtId);
      }
    int idx = outputPD->AddArray(eigenvalues);
    outputPD->SetActiveAttribute(idx, vtkDataSetAttributes::SCALARS);
    lambda->Delete();
    eigenvalues->Delete();

    // Remove data array named MaxEigenvalue automatically created
    // by vtkTensorGlyphSameEigensystem as that data array is comprised in
    // the vector data array named Eigenvalues provided by this class.
    if (outputPD->GetArray("MaxEigenvalue"))
      {
      outputPD->RemoveArray("MaxEigenvalue");
      }
    }
  numGlyphPts->Delete();
  append->Delete();

  return 1;
}


void vtkSuperquadricTensorGlyphFilter::PrintSelf(ostream& os, vtkIndent indent)
{
  this->Superclass::PrintSelf(os,indent);
}
